using System;
using System.Collections.Concurrent;
using System.Diagnostics.CodeAnalysis;
using System.Numerics;
using System.Threading;
using System.Threading.Tasks;

// ReSharper disable RedundantExtendsListEntry

namespace FastRng;

/// <summary>
/// A fast multi-threaded pseudo random number generator.
/// </summary>
/// <remarks>
/// Please note, that Math.NET's (https://www.mathdotnet.com/) random number generator is in some situations faster.
/// Unlike Math.NET, MultiThreadedRng is multithreaded. Consumers can use a token to cancel an operation.<br/><br/>
///
/// MultiThreadedRng using a shape fitter (a rejection sampler) to enforce arbitrary shapes of probabilities for
/// desired distributions. By using the shape fitter, it is even easier to define discontinuous, arbitrary functions
/// as shapes. Any consumer can define and use own distributions.<br/><br/>
/// 
/// This class uses the George Marsaglia's MWC algorithm. The algorithm's implementation is based loosely on John D.
/// Cook's (johndcook.com) implementation (https://www.codeproject.com/Articles/25172/Simple-Random-Number-Generation).
/// Thanks, John, for the inspiration.<br/><br/>
/// </remarks>
public sealed class MultiThreadedRng<TNum> : IRandom<TNum>, IDisposable where TNum : IFloatingPointIeee754<TNum>, IAdditionOperators<TNum, TNum, TNum>
{
#if DEBUG
    private const int BUFFER_SIZE = 1_000_000;
#else
    private const int BUFFER_SIZE = 1_000_000;
#endif

    // The queue size means, how many buffers we store in a queue at the same time:
    private const int QUEUE_SIZE_FLOAT = 2;
    
    // The queue size means, how many buffers we store in a queue at the same time:
    private const int QUEUE_SIZE_INT = QUEUE_SIZE_FLOAT * 2;

    private static readonly TNum CONST_FLOAT_CONVERSION = TNum.CreateChecked(2.328306435454494e-10f);
    
    // ReSharper disable StaticMemberInGenericType
    private static readonly object LOCKER = new();
    // ReSharper restore StaticMemberInGenericType
        
    // Gets used to stop the producer threads:
    private readonly CancellationTokenSource producerTokenSource = new();
        
    // The time a thread waits e.g. to check if the queue needs a new buffer:
    private readonly TimeSpan waiter = TimeSpan.FromMilliseconds(10);
        
    // The first queue, where to store buffers of random uint numbers:
    private readonly ConcurrentQueue<uint[]> queueIntegers = new();
        
    // The second queue, where to store buffers of uniform random floating point numbers:
    private readonly ConcurrentQueue<TNum[]> queueFloats = new();

    // The uint producer thread:
    private Thread producerRandomUint;
        
    // The uniform float producer thread:
    private Thread producerRandomUniformDistributedFloat;
    
    private readonly UIntChannelProducer independentUIntProducer = new();
        
    // Variable w and z for the uint generator. Both get used
    // as seeding variable as well (cf. constructors)
    private uint mW;
    private uint mZ;
        
    // This is the current buffer for the consumer side, i.e., the public interfaces:
    private TNum[] currentBuffer = [];
        
    // The current pointer to the next current buffer's address to read from:
    private int currentBufferPointer = BUFFER_SIZE;

    #region Constructors

    /// <summary>
    /// Creates a multi-threaded random number generator.
    /// </summary>
    /// <remarks>
    /// This constructor uses the user's current local time to derive the necessary parameters for the generator.
    /// Thus, the results depend on the time when the generator was created.
    /// </remarks>
    public MultiThreadedRng()
    {
        //
        // Initialize the mW and mZ by using
        // the system's time.
        //
        var now = DateTime.Now;
        var ticks = now.Ticks;
        this.mW = (uint) (ticks >> 16);
        this.mZ = (uint) (ticks % 4_294_967_296);
        this.StartProducerThreads();
    }

    /// <summary>
    /// Creates a multi-threaded random number generator.
    /// </summary>
    /// <remarks>
    /// A multi-threaded random number generator created by this constructor is
    /// deterministic. It's behaviour is not depending on the time of its creation.<br/><br/>
    ///
    /// <b>Please note:</b> Although the number generator and all distributions are deterministic,
    /// the behavior of the consuming application might be non-deterministic. This is possible if
    /// the application with multiple threads consumes the numbers. The scheduling of the threads
    /// is up to the operating system and might not be predictable. 
    /// </remarks>
    /// <param name="seedU">A seed value to generate a deterministic generator.</param>
    public MultiThreadedRng(uint seedU)
    {
        this.mW = seedU;
        this.mZ = 362_436_069;
        this.StartProducerThreads();
    }
        
    /// <summary>
    /// Creates a multi-threaded random number generator.
    /// </summary>
    /// <remarks>
    /// A multi-threaded random number generator created by this constructor is
    /// deterministic. It's behaviour is not depending on the time of its creation.<br/><br/>
    ///
    /// <b>Please note:</b> Although the number generator and all distributions are deterministic,
    /// the behavior of the consuming application might be non-deterministic. This is possible if
    /// the application with multiple threads consumes the numbers. The scheduling of the threads
    /// is up to the operating system and might not be predictable. 
    /// </remarks>
    /// <param name="seedU">The first seed value.</param>
    /// <param name="seedV">The second seed value.</param>
    public MultiThreadedRng(uint seedU, uint seedV)
    {
        this.mW = seedU;
        this.mZ = seedV;
        this.StartProducerThreads();
    }

    private void StartProducerThreads()
    {
        this.producerRandomUint = new Thread(() => this.RandomProducerUint(this.producerTokenSource.Token)) {IsBackground = true};
        this.producerRandomUint.Start();
        this.producerRandomUniformDistributedFloat = new Thread(() => this.RandomProducerUniformDistributedFloat(this.producerTokenSource.Token)) {IsBackground = true};
        this.producerRandomUniformDistributedFloat.Start();
    }

    #endregion

    #region Producers

    [ExcludeFromCodeCoverage]
    private async void RandomProducerUint(CancellationToken cancellationToken)
    {
        try
        {
            while (!cancellationToken.IsCancellationRequested)
            {
                // A local next buffer, which gets filled next:
                var nextBuffer = new uint[BUFFER_SIZE];
                    
                // Produce the necessary number of random uints:
                for (var n = 0; n < nextBuffer.Length && !cancellationToken.IsCancellationRequested; n++)
                {
                    this.mZ = 36_969 * (this.mZ & 65_535) + (this.mZ >> 16);
                    this.mW = 18_000 * (this.mW & 65_535) + (this.mW >> 16);
                    nextBuffer[n] = (this.mZ << 16) + this.mW;
                }

                // Inside this loop, we try to enqueue the produced buffer:
                while (!cancellationToken.IsCancellationRequested)
                {
                    try
                    {
                        // Ensure that we do not produce more buffers, as configured: 
                        if (this.queueIntegers.Count < QUEUE_SIZE_INT)
                        {
                            this.queueIntegers.Enqueue(nextBuffer);
                            break;
                        }

                        // The queue was full. Wait a moment and try it again:
                        await Task.Delay(this.waiter, cancellationToken);
                    }
                    catch (TaskCanceledException)
                    {
                        // The producers should be stopped:
                        return;
                    }
                }
            }
        }
        catch (OperationCanceledException)
        {
        }
    }
        
    [ExcludeFromCodeCoverage]
    private async void RandomProducerUniformDistributedFloat(CancellationToken cancellationToken)
    {
        try
        {
            while (!cancellationToken.IsCancellationRequested)
            {
                // A local source buffer of uints: 
                uint[] bufferSource;
                    
                // Try to get the next source buffer:
                while (!this.queueIntegers.TryDequeue(out bufferSource) && !cancellationToken.IsCancellationRequested)
                    await Task.Delay(this.waiter, cancellationToken);

                // Case: The producers should be stopped:
                if(bufferSource == null)
                    return;
                    
                // A local buffer to fill with uniform floats:
                var nextBuffer = new TNum[BUFFER_SIZE];
                    
                // Generate the necessary number of floats:
                for (var n = 0; n < nextBuffer.Length && !cancellationToken.IsCancellationRequested; n++)
                    nextBuffer[n] = (TNum.CreateChecked(bufferSource[n]) + TNum.One) * CONST_FLOAT_CONVERSION;

                // Inside this loop, we try to enqueue the generated buffer:
                while (!cancellationToken.IsCancellationRequested)
                {
                    try
                    {
                        // Ensure that the queue contains only the configured number of buffers:
                        if (this.queueFloats.Count < QUEUE_SIZE_FLOAT)
                        {
                            this.queueFloats.Enqueue(nextBuffer);
                            break;
                        }

                        // The queue was full. Wait a moment and try it again:
                        await Task.Delay(this.waiter, cancellationToken);
                    }
                    catch (TaskCanceledException)
                    {
                        return;
                    }
                }
            }
        }
        catch (OperationCanceledException)
        {
        }
    }

    #endregion

    #region Implementing interface
    
    /// <summary>
    /// Returns a uniform distributed pseudo-random number from the interval (0,1].
    /// This means, the result 0 is impossible, whereas 1 is possible.
    /// </summary>
    /// <remarks>
    /// This method is thread-safe. You can consume numbers from the same generator
    /// by using multiple threads at the same time. 
    /// </remarks>
    public TNum GetUniform(CancellationToken cancel = default)
    {
        if (cancel.IsCancellationRequested)
            return TNum.NaN;
        
        Start:
        
        // Check if we have to load the next buffer:
        if (this.currentBufferPointer >= BUFFER_SIZE)
        {
            // We have to get the next buffer from the queue. This is a critical
            // section, because we have to ensure, that only one thread at a time
            // can get the next buffer:
            lock (LOCKER)
            {
                // We might not the first thread, which has to get the next buffer.
                // When some other thread has already got the next buffer, the pointer
                // was already reset to zero. In this case, we start over again:
                if(this.currentBufferPointer < BUFFER_SIZE)
                    goto Start;
                
                while (!this.queueFloats.TryDequeue(out this.currentBuffer))
                {
                    if (cancel.IsCancellationRequested)
                        return TNum.NaN;
                
                    Thread.Sleep(TimeSpan.FromMilliseconds(6));
                }
                
                // Reset the pointer for the next thread or call:
                this.currentBufferPointer = 0;
            }
        }
        
        // Made a local copy of the current pointer:
        var myPointer = this.currentBufferPointer;
        
        // Issue #8: This might happen when another thread interrupted the current thread, and
        // the other thread has already updated the pointer. In this case, we start over again.
        if (myPointer >= BUFFER_SIZE)
            goto Start;
        
        // Increment the pointer for the next thread or call:
        var nextPointer = myPointer + 1;
        
        // Try to update the pointer without locking other threads:
        if (Interlocked.CompareExchange(ref this.currentBufferPointer, nextPointer, myPointer) != myPointer)
            goto Start;

        //
        // Case: Success. We updated the pointer and, thus, can use it to read the number.
        //
        return this.currentBuffer[myPointer];
    }

    /// <inheritdoc />
    public int GetUniformInt(int max, CancellationToken cancel = default)
    {
        var valueTask = this.independentUIntProducer.GetNextAsync(cancel);
        return (int) (valueTask.AsTask().Result % (uint) max);
    }

    /// <inheritdoc />
    public int GetUniformInt(int min, int max, CancellationToken cancel = default)
    {
        var valueTask = this.independentUIntProducer.GetNextAsync(cancel);
        return (int) (valueTask.AsTask().Result % (uint) (max - min) + (uint) min);
    }
        
    private void StopProducer() => this.producerTokenSource.Cancel();

    /// <summary>
    /// Disposes this generator. It is important to dispose a generator,
    /// when it is no longer needed. Otherwise, the background threads
    /// are still running.
    /// </summary>
    public void Dispose()
    {
        this.independentUIntProducer.Dispose();
        this.StopProducer();
    }

    #endregion
}