using System;
using System.Diagnostics.CodeAnalysis;
using System.Linq;
using FastRng;
using NUnit.Framework;

namespace FastRngTests.Distributions;

[ExcludeFromCodeCoverage]
public class InverseExponentialLa10
{
    [Test]
    [Category(TestCategories.COVER)]
    [Category(TestCategories.NORMAL)]
    public void TestExponentialDistribution01()
    {
        using var rng = new MultiThreadedRng<float>();
        var dist = new FastRng.Distributions.InverseExponentialLa10<float>(rng);
        var fqa = new FrequencyAnalysis();
            
        for (var n = 0; n < 100_000; n++)
            fqa.CountThis(dist.NextNumber());
            
        var result = fqa.NormalizeAndPlotEvents(TestContext.WriteLine);

        Assert.That(result[0], Is.EqualTo(0.0000501746820562f).Within(0.0003f));
        Assert.That(result[1], Is.EqualTo(0.0000554515994322f).Within(0.0003f));
        Assert.That(result[2], Is.EqualTo(0.0000612834950532f).Within(0.0003f));
            
        Assert.That(result[21], Is.EqualTo(0.00040973497898f).Within(0.00045f));
        Assert.That(result[22], Is.EqualTo(0.000452827182887f).Within(0.00050f));
        Assert.That(result[23], Is.EqualTo(0.000500451433441f).Within(0.0006f));
            
        Assert.That(result[50], Is.EqualTo(0.007446583070924f).Within(0.003f));
            
        Assert.That(result[75], Is.EqualTo(0.090717953289412f).Within(0.02f));
        Assert.That(result[85], Is.EqualTo(0.246596963941606f).Within(0.05f));
        Assert.That(result[90], Is.EqualTo(0.406569659740598f).Within(0.08f));
            
        Assert.That(result[97], Is.EqualTo(0.81873075307798f).Within(0.08f));
        Assert.That(result[98], Is.EqualTo(0.904837418035957f).Within(0.08f));
        Assert.That(result[99], Is.EqualTo(0.999999999999999f).Within(0.08f));
    }
        
    [Test]
    [Category(TestCategories.COVER)]
    [Category(TestCategories.NORMAL)]
    public void TestExponentialGeneratorWithRange01()
    {
        using var rng = new MultiThreadedRng<float>();
        var dist = new FastRng.Distributions.InverseExponentialLa10<float>(rng);
        var samples = new float[1_000];
        for (var n = 0; n < samples.Length; n++)
            samples[n] = dist.NextNumber(-1.0f, 1.0f);
            
        Assert.That(samples.Min(), Is.GreaterThanOrEqualTo(-1.0f), "Min out of range");
        Assert.That(samples.Max(), Is.LessThanOrEqualTo(1.0f), "Max out of range");
    }
        
    [Test]
    [Category(TestCategories.COVER)]
    [Category(TestCategories.NORMAL)]
    public void TestExponentialGeneratorWithRange02()
    {
        using var rng = new MultiThreadedRng<float>();
        var dist = new FastRng.Distributions.InverseExponentialLa10<float>(rng);
        var samples = new float[1_000];
        for (var n = 0; n < samples.Length; n++)
            samples[n] = dist.NextNumber(0.0f, 1.0f);
            
        Assert.That(samples.Min(), Is.GreaterThanOrEqualTo(0.0f), "Min is out of range");
        Assert.That(samples.Max(), Is.LessThanOrEqualTo(1.0f), "Max is out of range");
    }

    [Test]
    [Category(TestCategories.COVER)]
    [Category(TestCategories.NORMAL)]
    public void NoRandomNumberGenerator01()
    {
        Assert.Throws<ArgumentNullException>(() => new FastRng.Distributions.InverseExponentialLa10<float>(null));
    }
}